"""
WHAD Enhanced ShockBurst PHY utility functions.

This module provides some functions related to ESB PHY, allowing
channel to frequency conversion or basic PHY configurations.
"""
from enum import IntEnum

from whad.phy import PhysicalLayer, GFSKModulationScheme, Endianness
from whad.scapy.layers.esb import ESB_Hdr


class FieldsSize(IntEnum):
    '''
    Size of major ESB fields (in bytes).
    '''
    PREAMBLE_SIZE = 1
    ADDRESS_SIZE = 5
    HEADER_SIZE = 2
    CRC_SIZE = 2


def frequency_to_channel(frequency) -> int:
    """
    Converts a frequency (in Hz) to an Enhanced ShockBurst channel.
    """
    return int(frequency/1000000) - 2400

def channel_to_frequency(channel) -> int:
    """
    Converts an Enhanced ShockBurst channel into a frequency (in Hz).
    """
    return (2400 + channel) * 1000000

def decoding(pkt, configuration = None):
    """Decode a received packet depending on the current configuration.
    """
    size = (pkt[6] & 0b11111100) >> 2
    pkt_size = FieldsSize.PREAMBLE_SIZE + FieldsSize.CRC_SIZE + FieldsSize.ADDRESS_SIZE
    pkt_size += FieldsSize.HEADER_SIZE + size
    return pkt[:pkt_size]


PHYS = {
    "ESB-1M": PhysicalLayer(
                modulation=GFSKModulationScheme(deviation=170000),
                datarate=1000000,
                endianness=Endianness.BIG,
                frequency_range=(2400000000, 2500000000),
                maximum_packet_size=31,
                synchronization_word=b"\xAA",
                scapy_layer=ESB_Hdr,
                format_address_function=lambda address:bytes.fromhex(address.replace(":","")),
                frequency_to_channel_function=frequency_to_channel,
                channel_to_frequency_function=channel_to_frequency,
                encoding_function=None,
                decoding_function=decoding
            ),

    "ESB-2M": PhysicalLayer(
                modulation=GFSKModulationScheme(deviation=500000),
                datarate=2000000,
                endianness=Endianness.BIG,
                frequency_range=(2400000000, 2500000000),
                maximum_packet_size=31,
                synchronization_word=b"\xAA",
                scapy_layer=ESB_Hdr,
                format_address_function=lambda address:bytes.fromhex(address.replace(":","")),
                frequency_to_channel_function=frequency_to_channel,
                channel_to_frequency_function=channel_to_frequency,
                encoding_function=None,
                decoding_function=decoding
            )
}



"""This module provides a TCPSocketDevice class that can be used with a WhadDeviceConnector
to interact with a TCP socket connected to a Whad-enable device. This class implements
a TCP socket client that connects to a remote device through a TCP socket.

It also provides a dedicated connector to be used as a TCP socket server.

This class handles device connection, disconnection and read/write operations. All the
parsing magic is performed in our WhadDevice class.
"""
import logging
import socket
import select
import re

from ipaddress import ip_address

from whad.exceptions import WhadDeviceNotReady, WhadDeviceDisconnected, WhadDeviceNotFound
from whad.hub.message import HubMessage

from .device import Device
from .connector import Connector

logger = logging.getLogger(__name__)

def is_valid_hostname(hostname: str) -> bool:
    """Determine if a hostname is valid.

    :param hostname: Hostname to check
    :type hostname: str
    :return: `True` if hostname is valid, `False` otherwise.
    :rtype: bool
    """
    if hostname[-1] == ".":
        # strip exactly one dot from the right, if present
        hostname = hostname[:-1]
    if len(hostname) > 253:
        return False

    labels = hostname.split(".")

    # the TLD must be not all-numeric
    if re.match(r"[0-9]+$", labels[-1]):
        return False

    allowed = re.compile(r"(?!-)[a-z0-9-]{1,63}(?<!-)$", re.IGNORECASE)
    return all(allowed.match(label) for label in labels)

class TcpSocket(Device):
    """
    TCP client class.
    """

    INTERFACE_NAME = "tcp"

    @classmethod
    def list(cls):
        '''
        Returns a list of available TCP socket devices.
        '''
        return None

    @classmethod
    def check_interface(cls, interface):
        '''
        This method checks dynamically if the provided interface can be instantiated.
        '''
        logger.info("Checking interface: %s", str(interface))
        if ":" in interface:
            host, port = interface.split(":")
            try:
                port = int(port)
            except ValueError:
                return False
        else:
            host = interface
            port = 12345

        try:
            ip_address(host)
        except ValueError:
            # It is not a valid IP address, is it a valid hostname ?
            if not is_valid_hostname(host):
                return False

        # We survived to previous code, let's check if we can establish a connection
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(1)
        try:
            return s.connect_ex((host, port)) == 0
        except:
            return False

    def __init__(self, interface):
        """
        Create device connection
        """
        super().__init__()

        # Connect to target TCP Socket device in non-blocking mode
        try:
            if isinstance(interface, tuple) and len(interface) == 2:
                self.__address = interface[0]
                self.__port = interface[1]
            elif isinstance(interface, str):
                if ":" in interface:
                    self.__address = interface.split(":")[0]
                    self.__port = int(interface.split(":")[1])
                else:
                    self.__address = interface
                    self.__port = 12345
            else:
                self.__address = None
                self.__port = None
        except ValueError:
            self.__address = None
            self.__port = None
        self.__socket = None
        self.__fileno = None
        self.__opened = False

    @property
    def identifier(self):
        '''
        Returns the identifier of the device (e.g., socket path).
        '''
        return str(self.__address) + ":" + str(self.__port)


    def open(self):
        """
        Open device.
        """
        if not self.__opened:
            try:
                # Open TCP socket
                self.__socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self.__socket.connect((self.__address, self.__port))
                self.__fileno = self.__socket.fileno()
                self.__opened = True
                # Ask parent class to run a background I/O thread
                super().open()
            except ConnectionRefusedError as conn_refused:
                logger.error('Connection refused')
                raise WhadDeviceNotFound from conn_refused

    def reset(self):
        """Reset device.

        This method is not supported by this type of device.
        """


    def close(self):
        """
        Close current device.
        """
        # Close underlying device.
        if self.__socket is not None:
            self.__socket.close()

        # Clear fileno and status
        self.__fileno = None
        self.__opened = False

        # Ask parent class to stop I/O thread
        super().close()


    def write(self, payload):
        """Writes data to the device. It relies on select() in order to make sure
        we are allowed to write to the device and wait without eating too much CPU
        if the device is not ready to be written to.

        :param bytes payload: Data to write
        :returns: number of bytes written to the device
        """
        logger.debug("sending data to TCP socket: %s", data.hex())
        if not self.__opened:
            raise WhadDeviceNotReady()

        nb_bytes_written = 0
        wlist = [self.__fileno]
        elist = [self.__fileno]
        _, writers, __ = select.select(
            [],
            wlist,
            elist
        )

        if len(writers) > 0:
            nb_bytes_written = self.__socket.send(payload)
        return nb_bytes_written

    def read(self) -> bytes:
        """Fetches data from the device, if there is any data to read. We call select()
        to make sure data is waiting to be read before reading it. Data is then sent to
        our parsing method through on_data_received() that will handle data reassembling
        and message parsing and dispatch.
        """
        try:
            if not self.__opened:
                raise WhadDeviceNotReady()

            rlist = [self.__fileno]
            wlist = []
            elist = [self.__fileno]

            readers, _, __ = select.select(
                rlist,
                wlist,
                elist,
                1
            )

            # Handle incoming messages if any
            if len(readers) > 0:
                data = self.__socket.recv(1024)
                if len(data) > 0:
                    return data
                else:
                    #logger.error('No data received from device')
                    raise WhadDeviceDisconnected()
        except ConnectionResetError:
            logger.error('Connection reset by peer')
        except Exception as exc:
            raise WhadDeviceDisconnected() from exc

    def change_transport_speed(self, speed):
        """Not supported by TCP socket devices.
        """


class TCPSocketConnector(Connector):
    """TCP socket server connector.

    This connector will create a TCP socket and accept a single connection
    to this socket. It will then forward every message receveived from the
    underlying WHAD device to the connected client, and the messages received
    from the client to the device (passthrough mode).
    """

    def __init__(self, device, address="127.0.0.1", port=12345):
        """Create a TCPSocketConnector
        """
        # No client connected
        self.__client = None

        # Input pipes
        self.__inpipe = bytearray()

        # Socket
        self.__socket = None

        # Shutdown require (to force exit in  ̀`serve()`)
        self.__shutdown_required = False

        # Initialize parent class (device connector)
        super().__init__(device)

        # Create our TCP socket
        self.__address = address
        self.__port = port

        logger.debug("TCP socket address and port: %s (%s)", str(self.__address),
                     str(self.__port))

    def on_data_received(self, data: bytes):
        """Handle incoming data and parse it.

        :param data: Incoming data
        :type data: bytes
        """
        logger.debug("received raw data from socket: %s", data.hex())
        self.__inpipe.extend(data)
        while len(self.__inpipe) > 2:
            # Is the magic correct ?
            if self.__inpipe[0] == 0xAC and self.__inpipe[1] == 0xBE:
                # Have we received a complete message ?
                if len(self.__inpipe) > 4:
                    msg_size = self.__inpipe[2] | (self.__inpipe[3] << 8)
                    if len(self.__inpipe) >= (msg_size+4):
                        raw_message = self.__inpipe[4:4+msg_size]

                        # Parse our message with our Protocol Hub
                        _msg = self.hub.parse(bytes(raw_message))

                        # Send to device
                        if _msg is not None:
                            logger.debug(("WHAD message successfully parsed, "
                                        "forward to underlying device"))
                            self.device.send_message(_msg)

                            # Notify message
                            self.on_msg_sent(_msg)

                        # Chomp
                        self.__inpipe = self.__inpipe[msg_size + 4:]
                    else:
                        break
                else:
                    break
            else:
                # Nope, that's not a header
                while len(self.__inpipe) >= 2:
                    if (self.__inpipe[0] != 0xAC) or (self.__inpipe[1] != 0xBE):
                        self.__inpipe = self.__inpipe[1:]
                    else:
                        break

    def shutdown(self):
        """Shutdown TCP connection.
        """
        self.__shutdown_required = True

    def serve(self, timeout=None):
        """Serve TCP socket

        Forward data received through our TCP socket to the underlying device.

        :param float timeout: number of seconds to wait for incoming data
        """
        logger.debug('Creating TCP socket server')
        self.__socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.__socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.__socket.bind((self.__address, self.__port))
        self.__socket.listen()

        logger.debug("Waiting for TCP socket connection on %s (%s)",
                     self.__address, self.__port)

        try:
            while not self.__shutdown_required:
                self.__client,_ = self.__socket.accept()
                self.device.open()
                try:
                    while not self.__shutdown_required:
                        rlist = [self.__client.fileno()]
                        readers, _, errors = select.select(rlist, [], rlist, timeout)

                        # Do we have pending data ?
                        if len(readers) > 0:
                            data = self.__client.recv(4096)
                            # Is socket closed ?
                            if len(data) == 0:
                                # Exit serve loop
                                break

                            # Process received data
                            self.on_data_received(data)
                        elif len(errors) > 0:
                            logger.debug('Error detected on server socket, exiting.')
                            break
                except ConnectionResetError:
                    # Reset the underlying device
                    self.device.reset()


                self.__client = None

        except BrokenPipeError:
            logger.error('Broken pipe.')
            self.__client = None


    # Message callbacks
    def on_any_msg(self, message):
        """Callback function to process incoming discovery messages.

        This method MUST be overriden by inherited classes.

        :param message: Discovery message
        """
        try:
            logger.debug('Received a message from device, forward to client if any')
            if self.__client is not None:
                # Convert message into bytes
                raw_message = message.serialize()

                # Define header
                header = [
                    0xAC, 0xBE,
                    len(raw_message) & 0xff,
                    (len(raw_message) >> 8) & 0xff
                ]

                # Send header followed by serialized message
                self.__client.send(bytes(header))
                self.__client.send(raw_message)
                logger.debug('Message sent to client')
        except BrokenPipeError:
            logger.debug('Client socket disconnected')

    def on_packet(self, packet):
        """Incoming packet handler.
        """

    def on_event(self, event):
        """Generic event handler.
        """

    def on_msg_sent(self, message):
        """Called when a message has successfully been sent.
        """

    def on_generic_msg(self, message: HubMessage):
        """Callback function to process incoming generic messages.

        This method MUST be overriden by inherited classes.

        :param message: Generic message
        :type message: HubMessage
        """

    def on_discovery_msg(self, message: HubMessage):
        """Callback method to process incoming discovery messages.

        :param message: Discovery message
        :type message: HubMessage
        """

    def on_domain_msg(self, domain, message: HubMessage):
        """Callback function to process incoming domain-related messages.

        This method MUST be overriden by inherited classes.

        :param message: Domain message
        """



"""Logitech Unifying scanning tool

This utility finds any Logitech Unifying device and displays captured packets,
with a quick interpretation. It allows the user to quickly identify a mouse or
a keyboard based on the intepreted values.
"""
import logging

# Required to display colored texts
from prompt_toolkit import HTML, print_formatted_text

# Scapy
from scapy.packet import Packet

# Whad dependencies
from whad.cli.app import CommandLineDeviceSink, run_app
from whad.unifying.connector import Sniffer, ESBAddress
from whad.esb.exceptions import InvalidESBAddressException
from whad.scapy.layers.unifying import Logitech_Unifying_Hdr, \
    Logitech_Keepalive_Payload, Logitech_Mouse_Payload, \
    Logitech_Encrypted_Keystroke_Payload, Logitech_Unencrypted_Keystroke_Payload

# Logging
logger = logging.getLogger(__name__)

TOOL_DESCRIPTION="""This tool scans for Logitech Unifying devices and tries to classify them
based on the data they send. It can search for devices in automatic mode,
listen on a specific channel or target a specific device.
"""

class UniScanApp(CommandLineDeviceSink):
    """
    Unifying scanning application.

    This class inherits from `CommandLineDeviceSink` as it cannot be chained
    with another WHAD tool and will consume the data sent by a compatible
    device.
    """

    def __init__(self):
        """Application uses an interface and has no commands.
        """
        super().__init__(
            description=TOOL_DESCRIPTION,
            interface=True,
            commands=False
        )

        # Channel option
        self.add_argument(
            '-c',
            '--channel',
            metavar='CHANNEL',
            dest='channel',
            type=int,
            default=None,
            help="Find devices on channel CHANNEL"
        )

        # Target device address option
        self.add_argument(
            '-a',
            '--address',
            metavar='ADDRESS',
            dest='address',
            default=None,
            help="Follow a specific device with address ADDRESS"
        )


    def run(self):
        """Override App's run() method to handle scripting feature.
        """
        try:
            # Launch pre-run tasks
            self.pre_run()

            # We need to have an interface specified
            if self.interface is not None:
                # If an address is not provided, we scan for devices
                if self.args.address is None:
                    # Start scanning
                    self.scan()
                else:
                    # If an address is provided, we check its format and start
                    # sniffing packets from this device
                    try:
                        addr = ESBAddress(self.args.address)
                        self.sniff(addr)
                    except InvalidESBAddressException:
                        # Invalid device address
                        self.error('Target address does not match the expected format !')

            else:
                # Missing interface.
                self.error('You need to specify an interface with option --interface.')

        except KeyboardInterrupt:
            self.warning('wuni-scan stopped (CTL-C)')

        # Launch post-run tasks
        self.post_run()

    def sniff(self, address: ESBAddress):
        """Sniff packets set to the device identified by the provided Unifying
        address (5 bytes).
        """
        # Create our sniffer
        sniffer = Sniffer(self.interface)

        # Configure the target device address and channel. If channel is None,
        # the hardware will loop over the 100 possible channels.
        channel = self.args.channel
        if channel is not None and (channel < 0 or  channel > 100):
            self.error(f'Invalid channel value ({channel}). Channel must be in the 0-100 range.')

        # Start sniffing
        sniffer.channel = channel
        sniffer.address = address.value
        sniffer.start()

        # If no channel is specified, then automatically loop on all channels.
        if self.args.channel is None:
            print_formatted_text(HTML(
                'Following device <ansicyan>{address}</ansicyan> in auto mode ...'
            ).format(address=str(address)))
        else:
            print_formatted_text(HTML((
                "Sniffing device <ansicyan>{address}</ansicyan> "
                "on channel <ansicyan>{channel}</ansicyan>..."
                )).format(address=str(address), channel=self.args.channel))

        # Loop while CTL-C is not pressed, wait for a Unifying packet and display
        # it with a specific formatting.
        while True:
            for packet in sniffer.sniff():
                if Logitech_Unifying_Hdr in packet:
                    self.show_packet(packet)

    def scan(self):
        """Search for Logitech Unifying devices based on packets they send.
        If channel is specified, will lock on this channel and detect devices,
        otherwise the hardware will loop on all channels and will report any
        Unifying packet captured.
        """
        # Create our sniffer
        sniffer = Sniffer(self.interface)

        # Enable sniffing on our channel
        channel = self.args.channel
        if channel is not None and (channel < 0 or channel > 100):
            self.error(f'Invalid channel value ({channel}). Channel must be in the 0-100 range.')
            return
        sniffer.channel = channel
        sniffer.start()

        # Sniff packets and display them
        if channel is not None:
            print(f"Scanning for Unifying devices on channel {channel}")
        else:
            print("Scanning for Unifying devices on channels 0-100 ...")

        for packet in sniffer.sniff():
            if Logitech_Unifying_Hdr in packet:
                self.show_packet(packet)

    def show_packet(self, packet: Packet):
        """Display a Logitech Unifying packet with a short analysis.
        """
        # Extract Logitech Unifying payload
        payload = packet[Logitech_Unifying_Hdr]

        # Determine device type and generate comment
        comment = ''
        if Logitech_Mouse_Payload in packet:
            # Unencrypted mouse move, decode
            comment = '| <b>Mouse</b> (movement)'
        elif Logitech_Encrypted_Keystroke_Payload in packet:
            # Encrypted keystroke
            comment = '| <b>Keyboard</b> (encrypted keystroke)'
        elif Logitech_Unencrypted_Keystroke_Payload in packet:
            # Unencrypted keystroke
            comment = '| <b>Keyboard</b> <ansired>(unencrypted keystroke)</ansired>'
        elif Logitech_Keepalive_Payload in packet:
            # Unifying keep-alive
            comment = '| keep-alive'

        print_formatted_text(HTML((
            "<ansicyan>[{channel:03d}]</ansicyan><ansicyan>[{addr}]</ansicyan> "
            "<b>{payload}</b> %s" % comment
        )).format(channel=packet.metadata.channel,
                    addr=str(
                        packet.address if
                        hasattr(packet, "address") else
                        packet.metadata.address
                    ),
                    payload=bytes(payload).hex()))


def wuni_scan_main():
    """Logitech Unifying scanner main routine.
    """
    app = UniScanApp()
    run_app(app)



from enum import IntEnum
"""
Constants implemented as Enum for Zigbee APL layer.
"""

class LogicalDeviceType(IntEnum):
    """
    Represents the logical device type of a node.
    """
    COORDINATOR = 0
    ROUTER = 1
    END_DEVICE = 2

# Human Readable string for Zigbee Profile IDs
ZIGBEE_PROFILE_IDENTIFIERS = {
    0x0101 : "Industrial plant monitoring",
    0x0104 : "Home Automation",
    0x0105 : "Commercial Building Automation",
    0x0107 : "Telecom Applications",
    0x0108 : "Personal Home and Hospital Care",
    0x0109 : "Advanced Metering Initiative",
}

# Human Readable string for Zigbee Device IDs
ZIGBEE_DEVICE_IDENTIFIERS = {
    0x0000 : "ON/OFF Switch",
    0x0001 : "Level Control Switch",
    0x0002 : "ON/OFF Output",
    0x0003 : "Level Controllable Output",
    0x0004 : "Scene Selector",
    0x0005 : "Configuration Tool",
    0x0006 : "Remote control",
    0x0007 : "Combined Interface",
    0x0008 : "Range Extender",
    0x0009 : "Mains Power Outlet",
    0x0100 : "ON/OFF Light",
    0x0101 : "Dimmable Light",
    0x0102 : "Color Dimmable Light",
    0x0103 : "ON/OFF Light Switch",
    0x0104 : "Dimmer Switch",
    0x0105 : "Color Dimmer Switch",
    0x0106 : "Light Sensor",
    0x0107 : "Occupancy Sensor",
    0x0200 : "Shade",
    0x0201 : "Shade Controller",
    0x0300 : "Heating/Cooling Unit",
    0x0301 : "Thermostat",
    0x0302 : "Temperature Sensor",
    0x0303 : "Pump",
    0x0304 : "Pump Controller",
    0x0305 : "Pressure Sensor",
    0x0306 : "Flow sensor",
    0x0400 : "IAS Control and Indicating Equipment",
    0x0401 : "IAS Ancillary Control Equipment",
    0x0402 : "IAS Zone",
    0x0403 : "IAS Warning Device",
}

# Human readable list of Zigbee Cluster IDs
ZIGBEE_CLUSTER_IDENTIFIERS = {
    0x0000 : ("Generic", "Basic"),
    0x0001 : ("Generic", "Power configuration"),
    0x0002 : ("Generic", "Device temperature configuration"),
    0x0003 : ("Generic", "Identify"),
    0x0004 : ("Generic", "Groups"),
    0x0005 : ("Generic", "Scenes"),
    0x0006 : ("Generic", "ON/OFF"),
    0x0007 : ("Generic", "ON/OFF Switch configuration"),
    0x0008 : ("Generic", "Level Control"),
    0x0009 : ("Generic", "Alarms"),
    0x000A : ("Generic", "Time"),
    0x000B : ("Generic", "RSSI Location"),
    0x0100 : ("Closures", "Shade Configuration"),
    0x0200 : ("HVAC", "Pump Configuration and Control"),
    0x0201 : ("HVAC","Thermostat"),
    0x0202 : ("HVAC","Fan control"),
    0x0203 : ("HVAC","Dehumidification Control"),
    0x0204 : ("HVAC","Thermostat User Interface Configuration"),
    0x0300 : ("Lighting","Color Control"),
    0x0301 : ("Lighting","Ballast Configuration"),
    0x0400 : ("Measurement and sensing","Luminance Measurement"),
    0x0401 : ("Measurement and sensing","Luminance Level Sensing"),
    0x0402 : ("Measurement and sensing","Temperature Measurement"),
    0x0403 : ("Measurement and sensing","Pressure Measurement"),
    0x0404 : ("Measurement and sensing","Flow Measurement"),
    0x0405 : ("Measurement and sensing","Relative Humidity Measurement"),
    0x0406 : ("Measurement and sensing","Occupancy Sensing"),
    0x0500 : ("Security and Safety","IAS Zone"),
    0x0501 : ("Security and Safety","IAS ACE"),
    0x0502 : ("Security and Safety","IAS WD"),
}



"""
WHAD exceptions
"""

class RequiredImplementation(Exception):
    """
    This exception is raised when a class does not provide any implementation
    for a specific method, usually in interface classes.
    """


# Device discovery exceptions

class UnsupportedDomain(Exception):
    """
    Exception raised when a requested domain is not supported by a WHAD
    interface.
    """
    def __init__(self, domain: str):
        super().__init__()
        self.__domain = domain

    @property
    def domain(self):
        """Return the requested (and unsuppoted) domain name.
        """
        return self.__domain

    def __str__(self):
        """String representation.
        """
        return f"UnsupportedDomain({self.__domain})"

    def __repr__(self):
        """Python representation.
        """
        return str(self)

class UnsupportedCapability(Exception):
    """
    Exception raised when a specific requested capability is not supported by
    a WHAD interface.
    """

    def __init__(self, capability):
        super().__init__()
        self.__capability = capability

    @property
    def capability(self):
        """Return the unsupported capability.
        """
        return self.__capability

    def __str__(self):
        return f"UnsupportedCapability({self.__capability})"

    def __repr__(self):
        return str(self)


# Device communication exceptions

class WhadDeviceDisconnected(Exception):
    """
    Exception raised when a WHAD interface has disconnected.
    """

class WhadDeviceNotReady(Exception):
    """
    Exception raised when a WHAD interface is present but does not respond
    as expected.
    """

class WhadDeviceNotFound(Exception):
    """
    Raised when a specified WHAD interface has not been found.
    """

class WhadDeviceAccessDenied(Exception):
    """
    Exception raised when a WHAD interface cannot be accessed due to wrong
    permissions.
    """

    def __init__(self, device_name):
        self.__device_name = device_name
        super().__init__()

    def __str__(self):
        return f"WhadDeviceAccessDenied({self.__device_name}) - missing udev rules"

class WhadDeviceUnsupportedOperation(Exception):
    """
    Exception is raised when a specific requested operation is not supported by
    the WHAD interface.
    """

    def __init__(self, operation, message):
        self.__operation = operation
        self.__message = message

    def __str__(self):
        return f"WhadDeviceUnsupportedOperation({self.__operation}) - {self.__message}"

    def __repr__(self):
        return str(self)

class WhadDeviceTimeout(Exception):
    """
    Excepton raised when a WHAD interface communication timed out.
    """

    def __init__(self, message):
        self.__message = message

    def __str__(self):
        return f"WhadDeviceTimeout({self.__message})"

class WhadDeviceError(Exception):
    """
    Exception is raised when an error occurred during WHAD interface management.
    """

    def __init__(self, message):
        self.__message = message

    def __str__(self):
        return f"WhadDeviceError({self.__message})"
    
    @property
    def message(self) -> str:
        return self.__message

# External tools exceptions

class ExternalToolNotFound(Exception):
    """
    Exception is raised when an external tool cannot be found.
    """

    def __init__(self, tool):
        super().__init__()
        self.__tool = tool

    def __str__(self):
        return f"ExternalToolNotFound({self.__tool})"

    def __repr__(self):
        return str(self)

# Triggers exception
class TriggerNotAssociated(Exception):
    """
    Exception is raised when a trigger cannot be found.
    """

class InvalidTriggerPattern(Exception):
    """
    Exception is raised when a provided trigger pattern is invalid.
    """

# Generic replay
class ReplayInvalidRole(Exception):
    """
    Exception raised when an invalid role has been selected regarding a requested
    operation.
    """

    def __repr__(self):
        return "ReplayInvalidRole"



"""
WHAD Enhanced ShockBurst

This module provides the `ESBAddress` class that represents
an Enhanced ShockBurst device address.
"""
import re
from whad.esb.exceptions import InvalidESBAddressException

class ESBAddress:
    """This class represents an Enhanced ShockBurst address.
    """

    def __init__(self, address):
        """Initialize ESB address
        """
        if isinstance(address, str):
            if re.match(r'^([0-9a-fA-F]{2}\:){4}[0-9a-fA-F]{2}$', address) is not None:
                self.__value = bytes.fromhex(address.replace(":",""))
            elif re.match('[0-9a-fA-F]{10}$', address) is not None:
                self.__value = bytes.fromhex(address)
            else:
                raise InvalidESBAddressException
        elif isinstance(address, bytes) and len(address) > 1 and len(address) <= 5:
            self.__value = address
        else:
            raise InvalidESBAddressException

    def __eq__(self, other):
        return self.__value == other.value

    def __str__(self):
        return ":".join([f"{b:02x}" for b in self.__value])

    def __repr__(self):
        return f"ESBAddress({self})"

    @property
    def value(self) -> bytes:
        """Return the address value

        :return: ESB address as byte buffer
        :rtype: bytes
        """
        return self.__value

    @property
    def base(self) -> str:
        """Return the 4 MSBytes of this ESB address as string

        :return: ESB base address
        :rtype: str
        """
        return ":".join([f"{b:02x}" % b for b in self.__value[:4]])

    @property
    def prefix(self) -> str:
        """Return the ESB address prefix

        :return: ESB address prefix
        :rtype: str
        """
        return f"{self.__value[4]:02x}"

    @staticmethod
    def from_bytes(esb_addr_bytes):
        """Convert a 5-byte array into a valid ESB address.

        :param bytes bd_addr_bytes: Enhanced ShockBurst address as a bytearray.
        :rtype: ESBAddress
        :return: An instance of ESBAddress representing the corresponding ESB address.
        """
        if len(esb_addr_bytes) == 5:
            hex_address = esb_addr_bytes.hex()
            address = ":".join([hex_address[i*2:(i+1)*2] for i in range(int(len(hex_address)/2))])
            return ESBAddress(address)

        # Raise an error (invalid address size)
        raise InvalidESBAddressException



from scapy.packet import Packet, bind_layers
from scapy.fields import XByteEnumField, ByteField, LEShortField, ByteEnumField, \
    XByteField, BitField, BitEnumField, LESignedShortField, StrFixedLenField
from struct import pack
from whad.scapy.layers.esb import ESB_Payload_Hdr, SBAddressField, \
    guess_payload_class_esb

class Microsoft_Hdr(Packet):
    name = "Microsoft Header"
    fields_desc = [
        XByteEnumField("device_class",None, {0x0a: "keyboard", 0x08: "mouse"}),
        XByteEnumField("packet_type",  None, {0x78: "keystroke"}),
        XByteEnumField("model_id",None, {}),
        ByteField("unknown", 0x01),
        XByteField("checksum", None),
        ByteEnumField("valid_checksum", None, {0:"invalid", 1:"valid"}),
        LEShortField("sequence_number", None)
    ]

    @classmethod
    def compute_checksum(cls, data, key, init):
        checksum = init
        for i in data:
            checksum ^= i
        return checksum

    def pre_dissect(self, s):
        key = bytes.fromhex(self.underlayer.underlayer.address.replace(":",""))[::-1]

        unciphered = s[:4]

        count = 0
        deciphered = b""
        for i in s[4:-1]:
            deciphered += bytes([i ^ key[count]])
            count = (count + 1) % len(key)

        for i in range(len(key)):
            checksum = Microsoft_Hdr.compute_checksum(unciphered + deciphered, key, key[i] ^ 0xFF)
            if checksum == s[-1]:
                return unciphered + bytes([s[-1], int(True)]) + deciphered

        checksum = Microsoft_Hdr.compute_checksum(unciphered + s[4:-1], key, 0xFF)
        if checksum == s[-1]:
            return unciphered + bytes([s[-1], int(True)]) + s[4:-1]

        return unciphered + bytes([s[-1], int(False)]) + s[4:-1]

class Microsoft_Keystroke_Payload(Packet):
    name = "Microsoft Keystroke Payload"
    fields_desc = [
        ByteField("unknown2", 0x43),
        BitField("right_padding", 0, 1),
        BitEnumField("right_alt", 0, 1, {1:"pressed", 0:"released"}),
        BitEnumField("right_shift", 0, 1, {1:"pressed", 0:"released"}),
        BitEnumField("right_ctrl", 0, 1, {1:"pressed", 0:"released"}),
        BitField("left_padding", 0, 1),
        BitEnumField("left_alt", 0, 1, {1:"pressed", 0:"released"}),
        BitEnumField("left_shift", 0, 1, {1:"pressed", 0:"released"}),
        BitEnumField("left_ctrl", 0, 1, {1:"pressed", 0:"released"}),
        StrFixedLenField('hid_data', '\0\0\0\0\0\0\0', length=7),

    ]

class Microsoft_Mouse_Payload(Packet):
    name = "Microsoft Mouse Payload"
    fields_desc = [
        LEShortField("unknown2", 0x0040),
        BitField("padding", 0, 4),
        BitEnumField("special_click", 0, 1, {1:"pressed", 0:"released"}),
        BitEnumField("middle_click", 0, 1, {1:"pressed", 0:"released"}),
        BitEnumField("right_click", 0, 1, {1:"pressed", 0:"released"}),
        BitEnumField("left_click", 0, 1, {1:"pressed", 0:"released"}),
        LESignedShortField("x", None),
        LESignedShortField("y", None),
        LESignedShortField("wheel_y", None),
        LESignedShortField("wheel_x", None),
        ByteField("unknown3", 0x01),

    ]

def guess_payload_class_microsoft(self, payload):
    if b"\x0f\x0f\x0f\x0f" == payload[:4]:
        return ESB_Ping_Request
    elif len(payload) == 0:
        return ESB_Ack_Response
    elif len(payload) >= 2 and payload[0] in (0x08, 0x0a) and payload[1] in (0x78, 0x38, 0x90):
        return Microsoft_Hdr
    else:
        return Packet.guess_payload_class(self, payload)


# Microsoft protocol

bind_layers(Microsoft_Hdr, Microsoft_Keystroke_Payload, packet_type=0x78)
bind_layers(Microsoft_Hdr, Microsoft_Mouse_Payload, packet_type=0x90)

def bind():
    ESB_Payload_Hdr.guess_payload_class = guess_payload_class_microsoft

def unbind():
    ESB_Payload_Hdr.guess_payload_class = guess_payload_class_esb



from whad.zigbee.stack.apl.zdo.descriptors import NodeDescriptor, SimpleDescriptor
from whad.zigbee.stack.aps.constants import APSDestinationAddressMode
from whad.zigbee.stack.apl.cluster import Cluster
from whad.zigbee.stack.apl.constants import LogicalDeviceType
from whad.dot15d4.stack.mac.constants import MACDeviceType, MACPowerSource
from scapy.layers.zigbee import  ZigbeeDeviceProfile
from whad.scapy.layers.zdp import ZDPDeviceAnnce, ZDPNodeDescReq, ZDPNodeDescRsp, ZDPNWKAddrReq, \
    ZDPIEEEAddrReq, ZDPIEEEAddrRsp, ZDPActiveEPReq, ZDPActiveEPRsp, ZDPSimpleDescReq, ZDPSimpleDescRsp

class ZDPCluster(Cluster):
    """
    This class represents a Zigbee Device Profile cluster.
    """
    def __init__(self, zdo_object, cluster_id):
        super().__init__(cluster_id=cluster_id)
        self.zdo_object = zdo_object

    def send_data(
                    self,
                    command,
                    transaction,
                    destination_address_mode=APSDestinationAddressMode.SHORT_ADDRESS_DST_ENDPOINT_PRESENT,
                    destination_address=0xFFFF,
                    use_network_key=True,
                    destination_endpoint=0
    ):
        """
        Method allowing to send ASDU related to cluster.
        """
        zdp_command =  ZigbeeDeviceProfile(trans_seqnum=transaction) / command
        super().send_data(
            zdp_command,
            destination_address_mode=destination_address_mode,
            destination_address=destination_address,
            use_network_key=use_network_key,
            destination_endpoint=destination_endpoint
        )

    def on_data(self, asdu, source_address, source_address_mode, security_status, link_quality):
        """
        Callback called when an ASDU related to the current cluster is received.
        """
        self.zdo_object.on_cluster_data(
            asdu,
            source_address,
            source_address_mode,
            security_status,
            link_quality
        )


class ZDONWKAddrReq(ZDPCluster):
    def __init__(self, zdo_object):
        super().__init__(zdo_object, cluster_id=0x0000)

    def send_data(self, address=0xfffd, request_type=1, start_index=0, transaction=0):
        command = ZDPNWKAddrReq(
            ieee_addr=address,
            request_type=request_type,
            start_index=start_index
        )
        super().send_data(
            command,
            transaction,
            destination_address_mode=APSDestinationAddressMode.SHORT_ADDRESS_DST_ENDPOINT_PRESENT,
            destination_address=address,
            use_network_key=True,
            destination_endpoint=0
        )

class ZDONWKAddrRsp(ZDPCluster):
    def __init__(self, zdo_object):
        super().__init__(zdo_object, cluster_id=0x8000)


class ZDOIEEEAddrReq(ZDPCluster):
    def __init__(self, zdo_object):
        super().__init__(zdo_object, cluster_id=0x0001)

    def send_data(self, address, request_type=1, start_index=0, transaction=0):
        command = ZDPIEEEAddrReq(
            nwk_addr=address,
            request_type=request_type,
            start_index=start_index
        )
        super().send_data(
            command,
            transaction,
            destination_address_mode=APSDestinationAddressMode.SHORT_ADDRESS_DST_ENDPOINT_PRESENT,
            destination_address=address,
            use_network_key=True,
            destination_endpoint=0
        )

class ZDOIEEEAddrRsp(ZDPCluster):
    def __init__(self, zdo_object):
        super().__init__(zdo_object, cluster_id=0x8001)


    def send_data(self, nwk_address, ieee_address,remote_address, status=0, num_assoc_dev=0, start_index=0, associated_devices=[], transaction=0):

        command = ZDPIEEEAddrRsp(
            ieee_addr=ieee_address,
            nwk_addr=nwk_address,
            status=status
        )
        if status == 0:
            command.num_assoc_dev=num_assoc_dev
            command.start_index = start_index
            command.associated_devices = associated_devices


        super().send_data(
            command,
            transaction,
            destination_address_mode=APSDestinationAddressMode.SHORT_ADDRESS_DST_ENDPOINT_PRESENT,
            destination_address=remote_address,
            use_network_key=True,
            destination_endpoint=0
        )
class ZDODeviceAnnce(ZDPCluster):
    def __init__(self, zdo_object):
        super().__init__(zdo_object, cluster_id=0x0013)

    def send_data(self, transaction=0):
        node_descriptor = self.application.configuration.get("configNodeDescriptor")
        nwk_layer = self.application.manager.get_layer('nwk')
        command = ZDPDeviceAnnce(
            nwk_addr = nwk_layer.database.get("nwkNetworkAddress"),
            ieee_addr = nwk_layer.database.get("nwkIeeeAddress"),
            allocate_address = int(node_descriptor.allocate_address),
            security_capability = int(node_descriptor.security_capability),
            receiver_on_when_idle = int(node_descriptor.receiver_on_when_idle),
            power_source = int(node_descriptor.power_source),
            device_type = int(node_descriptor.logical_type),
            alternate_pan_coordinator = int(node_descriptor.alternate_pan_coordinator)
        )
        super().send_data(
            command,
            transaction,
            destination_address_mode=APSDestinationAddressMode.SHORT_ADDRESS_DST_ENDPOINT_PRESENT,
            destination_address=0xfffd,
            use_network_key=True,
            destination_endpoint=0
        )


class ZDONodeDescReq(ZDPCluster):
    def __init__(self, zdo_object):
        super().__init__(zdo_object, cluster_id=0x0002)

    def send_data(self, address, transaction=0):
        command = ZDPNodeDescReq(
            nwk_addr = address
        )
        super().send_data(
            command,
            transaction,
            destination_address_mode=APSDestinationAddressMode.SHORT_ADDRESS_DST_ENDPOINT_PRESENT,
            destination_address=address,
            use_network_key=True,
            destination_endpoint=0
        )

class ZDONodeDescRsp(ZDPCluster):
    def __init__(self, zdo_object):
        super().__init__(zdo_object, cluster_id=0x8002)

    def send_data(self, address, status, node_descriptor=None, transaction=0):
        if status == 0:
            command =  ZDPNodeDescRsp(
                nwk_addr=address,
                status=status,
                logical_type=int(node_descriptor.logical_type),
                complex_descriptor_available=int(node_descriptor.complex_descriptor_available),
                user_descriptor_available=int(node_descriptor.user_descriptor_available),
                aps_flags=node_descriptor.aps_flags,
                support_868_mhz=int(node_descriptor.support_868_mhz),
                support_902_mhz=int(node_descriptor.support_902_mhz),
                support_2400_mhz=int(node_descriptor.support_2400_mhz),
                alternate_pan_coordinator=int(node_descriptor.alternate_pan_coordinator),
                device_type=int(node_descriptor.device_type),
                power_source=int(node_descriptor.power_source),
                receiver_on_when_idle=int(node_descriptor.receiver_on_when_idle),
                security_capability=int(node_descriptor.security_capability),
                allocate_address=int(node_descriptor.allocate_address),
                manufacturer_code=node_descriptor.manufacturer_code,
                max_buffer_size=node_descriptor.max_buffer_size,
                max_incoming_transfer_size=node_descriptor.max_incoming_transfer_size,
                server_primary_trust_center = int(node_descriptor.server_primary_trust_center),
                server_backup_trust_center = int(node_descriptor.server_backup_trust_center),
                server_primary_binding_table_cache = int(node_descriptor.server_primary_binding_table_cache),
                server_backup_binding_table_cache = int(node_descriptor.server_backup_binding_table_cache),
                server_primary_discovery_cache = int(node_descriptor.server_primary_discovery_cache),
                server_backup_discovery_cache = int(node_descriptor.server_backup_discovery_cache),
                network_manager = int(node_descriptor.network_manager),
                stack_compliance_revision = node_descriptor.stack_compliance_revision,
                max_outgoing_transfer_size = node_descriptor.max_outgoing_transfer_size,
                extended_active_endpoint_list_available = int(node_descriptor.extended_active_endpoint_list_available),
                extended_simple_descriptors_list_available = int(node_descriptor.extended_simple_descriptors_list_available)
            )
        else:
            command =  ZDPNodeDescRsp(
                nwk_addr=address,
                status=status
            )
        super().send_data(
            command,
            transaction,
            destination_address_mode=APSDestinationAddressMode.SHORT_ADDRESS_DST_ENDPOINT_PRESENT,
            destination_address=address,
            use_network_key=True,
            destination_endpoint=0
        )


class ZDOActiveEPReq(ZDPCluster):
    def __init__(self, zdo_object):
        super().__init__(zdo_object, cluster_id=0x0005)

    def send_data(self, address, transaction=0):
        command = ZDPActiveEPReq(
            nwk_addr = address
        )
        super().send_data(
            command,
            transaction,
            destination_address_mode=APSDestinationAddressMode.SHORT_ADDRESS_DST_ENDPOINT_PRESENT,
            destination_address=address,
            use_network_key=True,
            destination_endpoint=0
        )

class ZDOActiveEPRsp(ZDPCluster):
    def __init__(self, zdo_object):
        super().__init__(zdo_object, cluster_id=0x8005)


    def send_data(self, local_address, remote_address, status=0, endpoints=[], transaction=0):
        command = ZDPActiveEPRsp(
            status=status,
            nwk_addr = local_address,
            active_endpoints=endpoints,
            num_active_endpoints=len(endpoints)
        )

        super().send_data(
            command,
            transaction,
            destination_address_mode=APSDestinationAddressMode.SHORT_ADDRESS_DST_ENDPOINT_PRESENT,
            destination_address=remote_address,
            use_network_key=True,
            destination_endpoint=0
        )

class ZDOSimpleDescReq(ZDPCluster):
    def __init__(self, zdo_object):
        super().__init__(zdo_object, cluster_id=0x0004)

    def send_data(self, address, endpoint, transaction=0):
        command = ZDPSimpleDescReq(
            nwk_addr = address,
            endpoint = endpoint
        )
        super().send_data(
            command,
            transaction,
            destination_address_mode=APSDestinationAddressMode.SHORT_ADDRESS_DST_ENDPOINT_PRESENT,
            destination_address=address,
            use_network_key=True,
            destination_endpoint=0
        )

class ZDOSimpleDescRsp(ZDPCluster):
    def __init__(self, zdo_object):
        super().__init__(zdo_object, cluster_id=0x8004)

    def send_data(self, status, local_address, remote_address, descriptor=None, transaction=0):
        if descriptor is None:
            command = ZDPSimpleDescRsp(
                status = status,
                nwk_addr = address,
                descriptor_length = 0
            )
        else:
            command = ZDPSimpleDescRsp(
                status = status,
                nwk_addr = local_address,
                descriptor_length = (
                    6 +
                    (1 + 2 * len(descriptor.input_clusters)) +
                    (1 + 2 * len(descriptor.output_clusters))
                ),
                endpoint = descriptor.endpoint,
                profile_identifier = descriptor.profile_identifier,
                device_identifier = descriptor.device_identifier,
                device_version = descriptor.device_version,
                input_clusters_count = len(descriptor.input_clusters),
                input_clusters = descriptor.input_clusters,
                output_clusters_count = len(descriptor.output_clusters),
                output_clusters = descriptor.output_clusters
            )
        super().send_data(
            command,
            transaction,
            destination_address_mode=APSDestinationAddressMode.SHORT_ADDRESS_DST_ENDPOINT_PRESENT,
            destination_address=remote_address,
            use_network_key=True,
            destination_endpoint=0
        )
